/****************************************************************************

 This file is part of the qdu distribution
 Copyright (C) 2002, 2003, 2004  Gilles Debunne (Gilles.Debunne@imag.fr)
 Version 2.2-2, packaged on Tuesday February 1, 2005.

 http://artis.imag.fr/Membres/Gilles.Debunne/Code/QDU

 QDU is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 QDU is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with QDU; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/
#include "directoryView.h"

#include "math.h"

#include <stdlib.h>
#include <iostream>

#include <qapplication.h>
#include <qcursor.h>
#include <qdir.h>
#include <qmainwindow.h>
#include <qmessagebox.h>
#include <qpopupmenu.h>
#include <qregexp.h>
#include <qstatusbar.h>
#include <qtimer.h>

using namespace std;

DirectoryView::DirectoryView( QWidget *parent, const char *name)
  : QListView( parent, name )
{
  if (sizeof(QDU_UINT64) != 8)
    QMessageBox::critical(this, "QDU int error", "File sizes are not coded with 64 bits.<br>Large sizes may be clamped (usually to 2Gb)");

  // Create main window
  addColumn( "Name", 250 );
  addColumn( "Size" );
  addColumn( "" );

  // Sort by size
  setSorting( 1 );
  setResizeMode( QListView::LastColumn );
  setSelectionMode( QListView::Extended );
  setShowSortIndicator( true );
  
  connect( this, SIGNAL( selectionChanged() ),
	   this, SLOT( updateSelectedItemsList() ) );

  connect( this, SIGNAL( contextMenuRequested( QListViewItem *, const QPoint &, int ) ),
	   this, SLOT( contextMenu( QListViewItem*, const QPoint &, int ) ) );

  connect( this, SIGNAL( clicked( QListViewItem*, const QPoint &, int ) ),
	   this, SLOT( buttonPressed( QListViewItem*, const QPoint &, int) ) );

  // Dummy Garbage Item. Will receive deleted files.
  QString garbageDir("/tmp/qduGarbage-");
  garbageDir += QDir::home().dirName();
  garbage = new Item(new QListView(), garbageDir);
  garbage->setVisible(false);
  garbage->setSelectable(false);
  
  initTypeExtension();

  showHidden_ = false;
  minThreshold_ = 0;
  scale_ = 1;
  typeFilter_ = 0;

  // Drag 'n drop
  dropItem = NULL;
  autoOpenTimer = new QTimer( this );
  mousePressed = false;
  setAcceptDrops( true );
  viewport()->setAcceptDrops( true );
  connect( autoOpenTimer, SIGNAL( timeout() ), this, SLOT( openFolder() ) );

  // Undo
  undoIndex_ = 0;
  maxUndoIndex_= 0;
}

DirectoryView::~DirectoryView()
{
  delete autoOpenTimer;
  delete garbage->parent();
}

void DirectoryView::setRootDirectory(const QString& rootDir)
{
  Item* root = new Item(this, rootDir);
  root->rescanDisk();
}

void DirectoryView::goUpOneLevel()
{
  Item* root = (Item*)(firstChild());
  
  if (!root)
    return;

  QString newRootDir = (root->fullName).section('/', 0, -2);

  if (newRootDir.isEmpty())
    newRootDir = QDir::separator();
  
  if ((!QFileInfo(newRootDir).isReadable()) || (!QFileInfo(newRootDir).isDir()))
    {
      QMessageBox::warning(this, "Unreadable directory", "Unable to go up : directory " + newRootDir + " is unreadable", "Ok");
      return;
    }

  delete root;

  // Could (should ?) be optimized so that already scanned directory is not parsed again
  setRootDirectory(newRootDir);
  firstChild()->setOpen(true);
  fitScale();
}

void DirectoryView::refresh()
{
  if ((Item*)(firstChild()))
    ((Item*)(firstChild()))->rescanDisk();
}

void DirectoryView::openFolder()
{
  autoOpenTimer->stop();
  if ( dropItem && ((Item*)dropItem)->isDirectory && !dropItem->isOpen() )
    {
      dropItem->setOpen( true );
      dropItem->repaint();
    }
}

void DirectoryView::clearSelection()
{
  selectedItems_.clear();
  QListView::clearSelection();
  updateStatusBar();
}

void DirectoryView::initTypeExtension()
{
  Item::typeExtension[0]  = "";	// All
  Item::typeExtension[1]  = "h cpp C c cc ada ads adb p y java jar ms o ii rii a so"; // Code
  Item::typeExtension[2]  = "gif jpg tif tiff rgb rgbd bmp tga jpeg png icon xbm xpm sgi jfif idraw sho show rle bitmap fig inr"; // Images
  Item::typeExtension[3]  = "pdf"; // PDF
  Item::typeExtension[4]  = "eps ps ps epsi pstex"; // PS
  Item::typeExtension[5]  = "mp3 aiff aifc aif wav ra xam rdb au"; // Sounds
  Item::typeExtension[6]  = "mov mpg mpeg avi qt"; // Movies
  Item::typeExtension[7]  = "doc xls ppt psd ai wmf drw dot exe rtf"; // Windows
  Item::typeExtension[8]  = "tar tgz zip install"; // Archive
  Item::typeExtension[9]  = "tex tox ftex pstex_t aux bbl log blg bib sty cls bst dvi"; // Latex
  Item::typeExtension[10] = "wrl iv iff smf ccn scn 3ds m pfb pfa dxf mgf flt nff obj jcl mtl mb txl texel blend"; // 3D Models
  Item::typeExtension[11] = "html htm xml cgi php"; // Html
  Item::typeExtension[12] = "sh csh tcsh perl pl tcl el awk mel vim lex"; // Scripts
  Item::typeExtension[13] = "txt rtf"; // Texts
  Item::typeExtension[14] = "** core files"; // Core files
  Item::typeExtension[15] = "** All other extensions";

  for (unsigned short int i=1; i<14; ++i)
    {
      unsigned short int* value = new unsigned short int(i); 
      QStringList list = QStringList::split(' ', Item::typeExtension[i]);
      for ( QStringList::Iterator it = list.begin(); it != list.end(); ++it )
	Item::extType.insert(*it, value);
    }
}

void DirectoryView::setFileType(int i)
{
  typeFilter_ = i;
  ((Item*)firstChild())->invalidateSizeStrings();
  ((Item*)firstChild())->updateSize();
  ((Item*)firstChild())->updateVisibility();
  clearSelection();
}

// Max size, computed on currently visible items only. Used to fitScale
QDU_UINT64 DirectoryView::maxVisibleSize() const
{
  QDU_UINT64 max = 0;

  Item* item = (Item*)(itemAt( QPoint(0, 0) ));

  while (item)
    {
      if ((item->size > max) && (item->depth() > 0))
	max = item->size;
      item = (Item*)(item->itemBelow());
    }

  return max;
}

void DirectoryView::setScale(int s)
{
  scale_ = (unsigned int)(100.0 * exp(s/9.0));
  triggerUpdate();
}

void DirectoryView::fitScale()
{
  const QDU_UINT64 size = maxVisibleSize();
  unsigned int slider = 0;
  float s = 9.0 * log(size / (columnWidth(2)-treeStepSize()) / 100.0);
  if (s > 0)
    slider = 1+(int)s;
  else
    slider = 0;
  emit scaleChanged(slider);
  setScale(slider);
}

void DirectoryView::buttonPressed( QListViewItem* item, const QPoint &, int c )
{
  // A click its bar opens/closes the folder
  if (item && (c>0) && (((Item*)item)->isDirectory))
    item->setOpen(!item->isOpen());
}

void DirectoryView::keyPressEvent(QKeyEvent* e)
{
  switch (e->key())
    {
    case Qt::Key_Delete :
      if (e->state() == Qt::ShiftButton)
	deleteSelectedFiles();
      else
	if (garbageDirectoryIsOk())
	  applyToSelectedFiles(QDropEvent::Move, garbage, false);
      break;
    case Qt::Key_Z :
      if (e->state() == Qt::ControlButton)
	undo();
      break;
    case Qt::Key_Y :
      if (e->state() == Qt::ControlButton)
	redo();
      break;
    case Qt::Key_A :
      if (e->state() == Qt::ControlButton)
	selectAll(true);
      break;
    case Qt::Key_I :
      if (e->state() & Qt::ControlButton & Qt::ShiftButton)
	invertSelection();
      break;
    case Qt::Key_F1 :
      QMessageBox::information(this, "Qdu help",
			       "<center><h2>Qdu</h2><b>Graphical disk usage</b><br>Version 2.2 - Gilles Debunne</center><br>"
			       "Histogram display of file and directory sizes. Folders can be opened to display their contents. "
			       "Use right mouse button for context menu. Top buttons signification is displayed when mouse hovers.<br><br>"
			       "Select files with left button (use <b>Ctrl</b> and <b>Shift</b> for extended selection. "
			       "A directory can only be selected if none of its children are selected).<br><br>"
			       "Press <b>Delete</b> to delete the selected files and directories. "
			       "They are actually moved into <i>/tmp/qduGarbage</i> : press <b>Shift+Delete</b> to really delete (no undo).<br><br>"
			       "Drag and drop with <b>middle</b> mouse button to move selection to an other directory.<br><br>"
			       // "(press <b>Ctrl</b> to copy and <b>Ctrl+Shift</b> to create links).<br><br>"
			       "Use <b>Ctrl+Z</b> (resp. <b>Ctrl+Y</b>) to undo (resp. redo) your previous operations.", "Close");
      break;
    case Qt::Key_Escape :
      qApp->closeAllWindows();
      break;
    case Qt::Key_Q :
      if (e->state() == Qt::ControlButton)
	qApp->closeAllWindows();
      break;
    default:
      QListView::keyPressEvent(e);
    }
}

// fullName refers to the old name while text(0) is the new name.
// Uses applyToSelectedFiles to take advantage of the undo mechanism.
bool DirectoryView::renameItem(Item* item)
{
  clearSelection();
  selectedItems_.append(item);
  if (!(Item*)item->parent())
    QMessageBox::critical(this, "Internal error", "renameItem called on root item !");
  return applyToSelectedFiles(QDropEvent::Move, (Item*)item->parent());
}

void DirectoryView::deleteSelectedFiles()
{
  static QDir dummyDir;
  Item* it = selectedItems_.first();
  while (it)
    {
      Item* next = selectedItems_.next();

      bool success = dummyDir.remove(it->fullName);
      if (!success)
	{
	  QString command("rm -f ");
	  if (it->isDirectory)
	    command += "-r ";
	  QString fullName = it->fullName;
	  // Protects spaces
#if QT_VERSION >= 0x030100
	  command += fullName.replace(" ", "\\ ");
#else
	  command += fullName.replace(QRegExp(" "), "\\ ");
#endif
	  command += " >& /dev/null";
	  // cout << "rm command : " << command.latin1() << endl;
	  system(command.latin1());
	  // system return value is not reliable.
	  success = !(QFileInfo(it->fullName).exists());
	}

      if (success)
	{
	  it->addToParentSizes(false);
	  removeFromSelection(it);
	  delete it;
	}
      else
	if (next)
	  switch (QMessageBox::warning(this, "Error", "Unable to delete<br>" + it->fullName +
				       "<br>Check file and directory permissions.", "Continue", "Cancel"))
	    {
	    case 1 : return;
	    default: break;
	    }
	else
	  QMessageBox::warning(this, "Error", "Unable to delete<br>" + it->fullName, "Ok");

      it = next;
    }
}

// Check that directory exists and is writable.
bool DirectoryView::directoryIsWritable(const QString& dir)
{
  static QFileInfo dirInfo;
  dirInfo.setFile(dir);
  return (dirInfo.exists() && dirInfo.isDir() && dirInfo.isWritable());
}

// Check garbageDirectory is writable. Create a directory if needed,
// possibly trying new names until one is available.
// Returns true if garbageDirectory is ok.
bool DirectoryView::garbageDirectoryIsOk()
{
  if (directoryIsWritable(garbage->fullName))
    return true;
  else
    {
      // Try new dir names until it is ok
      static int count = 1;
      QFileInfo fileInfo(garbage->fullName);
      while (fileInfo.exists() && !(fileInfo.isDir() && fileInfo.isWritable()))
	{
	  (garbage->fullName).replace( QRegExp("-\\d+$"), "");
	  (garbage->fullName) += '-' + QString::number(count++);
	  fileInfo.setFile(garbage->fullName);
	}

      // dir name is available, but no directory exists : create one.
      if (!fileInfo.exists())
	{
	  QDir dummyDir;
	  if (dummyDir.mkdir(garbage->fullName))
	    QMessageBox::information(this, "New garbage directory created", 
				     "A new garbage directory has been created in<br>" + garbage->fullName +
				     "<br><br>You should remove it when you are done with qdu, or protect your privacy with a command like:<br>"
				     "<code>chmod 700 " + garbage->fullName +
				     "</code><br><br>Press <b>Shift+Delete</b> to really delete a selection (warning, no undo available).");
	  else
	    {
	      QMessageBox::warning(this, "Error", "Unable to create garbage directory<br>" + garbage->fullName);
	      return false;
	    }
	}

      return true;
    }
}

// Core function
// Applies action to all selectedItems_. Files and dirs are (moved/Copied/Linked) in destDir.
// When overWrite is true, new file names are found to avoir overwriting existing files (used for garbage).
bool DirectoryView::applyToSelectedFiles(QDropEvent::Action action, Item* destDir, bool overWrite)
{
  if (!directoryIsWritable(destDir->fullName))
    {
      QMessageBox::warning(this, "Unwritable directory", "Cannot write to directory<br>" + destDir->fullName);
      return false;
    }

  beginUndoEvent();
  
  Item* it = selectedItems_.first();
  while (it)
    {
      Item* next = selectedItems_.next();
      
      QString target = destDir->fullName + '/' + it->text(0);

      if ( it->fullName == target )
	{
	  it = next;
	  continue;
	}
      
      QString newName = it->text(0);
      // oldName and newName are identical, except when renameItem uses this procedure
      QString oldName = it->fullName.section('/', -1);
      QFileInfo fileInfo(target);

      // What to do when target already exists 
      bool overWriteThisFile = overWrite;
      if (fileInfo.exists())
	{
	  if (overWriteThisFile)
	    {
	      switch (QMessageBox::warning(this, "Name already exists",
					   target + "<br>already exists.",
					   "Rename ", "Overwrite", "Cancel"))
		{
		case 0 : overWriteThisFile = false; break;
		case 1 : break;
		case 2 : endUndoEvent(); return false;
		default: break;
		}
	    }
	  
	  if (!overWriteThisFile)
	    {
	      // Append a number to target name until a non existing name is found
	      int count = 1;
	      while (fileInfo.exists())
		{
		  newName = it->text(0) + "." + QString::number(count++);
		  target = destDir->fullName + '/' + newName;
		  fileInfo.setFile(target);
		}
	    }
	}

      Item* parent = (Item*)(it->parent());

      if (!parent)
	QMessageBox::critical(this, "Internal error", "Trying to " + actionString(action) + "root item !");
      
      if ( performAction(action, parent, it, destDir, newName) )
	{
	  addEventInUndoHistory(action, parent, it, destDir, oldName, newName);
	  removeFromSelection(it);
	}
      else
	if (next)
	  {
	    if (QMessageBox::warning(this, "Error", "Unable to " + actionString(action) + oldName + " to directory<br>" +
				     "Check file and directory permissions.<br>" +
				     destDir->fullName, "Continue", "Cancel") == 1)
	      {
		endUndoEvent();
		return false;
	      }
	  }
	else
	  QMessageBox::warning(this, "Error", "Unable to " + actionString(action) + oldName + " to directory<br>" +
			       destDir->fullName, "Ok");

      it = next;
    }
  endUndoEvent();
  return true;
}

// Update listView (children, sizes...)
bool DirectoryView::performAction(QDropEvent::Action action, Item* sourceDir, Item* item, Item* targetDir, const QString& newName)
{
  static QDir dummyDir;

  bool success = false;
  
  // Possible when redo/undo and file has really been deleted since then
  if (!sourceDir || !item || !targetDir)
    {
      qWarning("Internal error - Item or directory no longer exists in performAction");
      return false;
    }

  if (!QFileInfo(item->fullName).exists())
    QMessageBox::warning(this, "Error", item->fullName + " no longer exists.<br>"
			 "It has probably been deleted by an other application.<br>Consider rescaning directories", "Ok");
    
  if (!QFileInfo(targetDir->fullName).exists())
    QMessageBox::warning(this, "Error", "Directory " + targetDir->fullName + " no longer exists.<br>"
			 "It has probably been deleted by an other application.<br>Consider rescaning directories", "Ok");  

  const QString newFullName = targetDir->fullName + '/' + newName;
  
  if (action == QDropEvent::Move)
    {
      success = dummyDir.rename(item->fullName, newFullName);
      if (!success)
	{
	  QString command = QString("mv -f ") + item->fullName + " " + newFullName + " >& /dev/null";
	  // cout << command.latin1() << endl;
	  system(command.latin1());
	  // system return value is not reliable.
	  success = (!(QFileInfo(item->fullName).exists())) && ((QFileInfo(newFullName).exists()));
	}
    }
  else
    {
      // Copy and link : not yet implemented.
      // Item* newItem = new Item(targetDir, newName, item->isDirectory);
      // newItem->size = item->size;
      // QString command("cp ");
      // command += "cp ";
      // command += it->fullName + " " + target;
      // if ( system(command.latin1()) )

      // and so on. What about children ? Full copy ?
    }

  if (success)
    {
      if (action == QDropEvent::Move)
	{
	  item->addToParentSizes(false);
	  sourceDir->takeItem(item);
	  targetDir->insertItem(item);
	  item->addToParentSizes(true);
	}      

      // cout << "debug : " << action << " "  << item->fullName << "(" << item->text(0) << ") to " << newFullName << endl;

      item->setText(0, newName);
      item->fullName = newFullName;
    }
  
  return success;
}

// Called when selection changes. Updates selectedItems_.
void DirectoryView::updateSelectedItemsList()
{
  selectedItems_.clear();

  // Update selectedItems_ list. Items are depth() sorted.

#if QT_VERSION >= 0x030200
    for (QListViewItemIterator it( this, QListViewItemIterator::Selected ); it.current(); ++it)
#else
    for (QListViewItemIterator it( this ); it.current(); ++it)
      if (it.current()->isSelected())
#endif
	{
	  // item must be depth sorted. Let use the best sorting algorithm.
	  Item* item = (Item*)( it.current() );
	  selectedItems_.last();
	  while (selectedItems_.current() && (selectedItems_.current()->depth() < item->depth()))
	    selectedItems_.prev();
	  selectedItems_.insert(selectedItems_.at()+1, item);
	}

  // Enforce selection constraint : a selection cancels all parent selection
  QPtrListIterator<Item> sel( selectedItems_ );
  Item* item;
  while ( (item = sel.current()) != NULL )
    {
      Item* parent = (Item*)(item->parent());
      while ( parent )
	{
	  if (parent->isSelected())
	    removeFromSelection(parent);
	  parent = (Item*)(parent->parent());
	}
      ++sel;
    }

  onlyOneItemIsSelected_ = (selectedItems_.count() == 1);
  
  updateStatusBar();
}

void DirectoryView::updateStatusBar()
{
  QString message("");

  if (selectedItems_.isEmpty())
    message = "No selection. Press F1 for help.";
  else
    if (onlyOneItemIsSelected_)
      {
	Item* item = selectedItems_.first();
	QFileInfo fileInfo(item->fullName);
	static const int currentYear = QDate::currentDate().year();
	message = item->text(0) + ", " + Item::stringOfSize(item->size) + ". Last modified on ";
	QDateTime dtModif = fileInfo.lastModified();
    	if (dtModif.date().year() != currentYear)
	  message += dtModif.toString("dddd d MMMM yyyy, hh:mm");
	else
	  message += dtModif.toString("dddd d MMMM, hh:mm");

	/* Returns too strnage results. Does not seem exact.
	QDateTime dtRead = fileInfo.lastRead();
	if (abs(dtRead.secsTo(dtModif)) > 60)
	  {
	    message += " (last accessed ";
	    if (dtRead.date() != dtModif.date())
	      if (dtRead.date().year() != currentYear)
		message += "on " + dtRead.toString("dddd d MMMM yyyy, hh:mm");
	      else
		message += "on " + dtRead.toString("dddd d MMMM, hh:mm");
	    else
	      message += "at " + dtRead.toString("hh:mm");
	    message += ").";
	  }
	*/
      }
    else
      {
	unsigned int nbFiles = 0;
	unsigned int nbDirs = 0;
	QDU_UINT64 size = 0;
	for ( Item* it = selectedItems_.first(); it; it = selectedItems_.next() )
	  {
	    if (it->isDirectory)
	      nbDirs++;
	    else
	      nbFiles++;
	    size += it->size;
	  }
	if (nbFiles)
	  message = QString::number(nbFiles) + " file" + ((nbFiles==1)?"":"s");
	if (nbFiles && nbDirs)
	  message += " and ";
	if (nbDirs)
	  message += QString::number(nbDirs) + " director" + ((nbDirs==1)?"y":"ies");
	message += " selected : " + Item::stringOfSize(size);
      }
  
  ((QMainWindow*)(parent()->parent()))->statusBar()->message(message);
}

void DirectoryView::contextMenu( QListViewItem *item, const QPoint &, int col )
{
  QPopupMenu menu( this );

  int GO_UP, RESCAN, RENAME, DELETE, SELECT_ALL, INVERT_SELECT;
  GO_UP = RESCAN = RENAME = DELETE = SELECT_ALL = INVERT_SELECT = -2;

  if ( item && col==0 )
    {
      if (onlyOneItemIsSelected_)
	if (((Item*)item)->isDirectory)
	  {
	    if (item->depth() == 0)
	      GO_UP = menu.insertItem( "Scan parent directory" );
	    DELETE = menu.insertItem( "Delete directory" );
	    RESCAN = menu.insertItem( "Rescan directory" );
	    RENAME = menu.insertItem( "Rename directory" );
	  }
	else
	  {
	    DELETE = menu.insertItem( "Delete file" );
	    RENAME = menu.insertItem( "Rename file" );
	  }
      else
	DELETE = menu.insertItem( "Delete selection" );
    }

  menu.insertSeparator();
  
  SELECT_ALL = menu.insertItem( "Select All" );
  INVERT_SELECT = menu.insertItem( "Invert selection" );

  // menu.insertSeparator();
  // int ADD_SELECTION    = menu.insertItem( "+ files" );
  // int REMOVE_SELECTION = menu.insertItem( "- files" );
  
  menu.setMouseTracking( true );
  int select = menu.exec( QCursor::pos() );
  
  if ( select == RESCAN )
    ((Item*)item)->rescanDisk();
  else
    if (( select == RENAME ) && (item->renameEnabled(0) ))
      item->startRename(0);
    else
      if ( ( select == DELETE ) && (garbageDirectoryIsOk()) )
	applyToSelectedFiles(QDropEvent::Move, garbage, false);
      else
	if ( select == SELECT_ALL )
	  selectAll(true);
	else
	  if ( select == INVERT_SELECT )
	    invertSelection();
	  else
	    if ( select == GO_UP )
	      goUpOneLevel();
}

void DirectoryView::removeFromSelection(Item* item)
{
  item->setSelected(false);
  selectedItems_.remove(item);
}

void DirectoryView::hideSelection(bool hide)
{
  Item* item;
  for ( QPtrListIterator<Item> sel( selectedItems_ ); (item = sel.current()) != NULL; ++sel )
    item->setSelected(!hide);
  
  triggerUpdate();
}

void DirectoryView::sliderMinThreshold(int m)
{
  minThreshold_ = (QDU_UINT64)(m*m*m/1000000.0 * ((Item*)firstChild())->size);
  ((Item*)firstChild())->updateVisibility();
}

void DirectoryView::toggleHidden(bool b)
{
  showHidden_ = b;
  ((Item*)firstChild())->updateVisibility();
}

QString DirectoryView::actionString(const QDropEvent::Action& action)
{
  switch (action)
    {
    case QDropEvent::Move : return QString("move "); break;
    case QDropEvent::Copy : return QString("copy "); break;
    case QDropEvent::Link : return QString("link "); break;
    default : return QString::null; break;
    }
}
